#include "system.h"

#include <inttypes.h>
#include <emmintrin.h>
#include <printf.h>
#include <stdarg.h>
#include <arpa/inet.h>

#include "debug.h"

#define	USE_EXAMPLE_IPADDR	0
#define	USE_EXAMPLE_FASTA	0
#define	USE_EXAMPLE_WIDGET	0
#define	USE_EXAMPLE_XMM		0
#define	USE_EXAMPLE_HEX		1

static int _tpf_debug;

#if	USE_EXAMPLE_XMM
static int pa_xmm;
static int mod_v4f;
static int mod_v2d;
static int mod_v16i;
static int mod_v8i;
static int mod_v4i;
static int mod_v2i;
#endif
#if	USE_EXAMPLE_HEX
static int pa_hex;
static int mod_phh;
static int mod_hC;
static int mod_hD;
static int mod_hN;
#endif

#ifdef	REFERENCE
struct printf_info {
    int prec;			/* Precision.  */
    int width;			/* Width.  */
    wchar_t spec;		/* Format letter.  */
    unsigned int is_long_double:1;	/* L flag.  */
    unsigned int is_short:1;	/* h flag.  */
    unsigned int is_long:1;	/* l flag.  */
    unsigned int alt:1;		/* # flag.  */
    unsigned int space:1;	/* Space flag.  */
    unsigned int left:1;	/* - flag.  */
    unsigned int showsign:1;	/* + flag.  */
    unsigned int group:1;	/* ' flag.  */
    unsigned int extra:1;	/* For special use.  */
    unsigned int is_char:1;	/* hh flag.  */
    unsigned int wide:1;	/* Nonzero for wide character streams.  */
    unsigned int i18n:1;	/* I flag.  */
    unsigned int __pad:4;	/* Unused so far.  */
    unsigned short int user;	/* Bits for user-installed modifiers.  */
    wchar_t pad;		/* Padding character.  */
};
#endif

/* XXX this needs %02X padding/capitalization. */
static char *itoa(int value, char *result, int base)
{
    /* check that the base if valid */
    if (base < 2 || base > 36) {
	*result = '\0';
	return result;
    }

    char *ptr = result;
    char *ptr1 = result;
    char tmp_char;
    int tmp_value;

    do {
	tmp_value = value;
	value /= base;
	*ptr++ =
	    "zyxwvutsrqponmlkjihgfedcba9876543210123456789abcdefghijklmnopqrstuvwxyz"
	    [35 + (tmp_value - value * base)];
    } while (value);

    /* Apply negative sign */
    if (tmp_value < 0)
	*ptr++ = '-';
    *ptr-- = '\0';
    while (ptr1 < ptr) {
	tmp_char = *ptr;
	*ptr-- = *ptr1;
	*ptr1++ = tmp_char;
    }
    return result;
}

static char * info2fmt(const struct printf_info * info, char *b)
{
    char *be = b;

    *be++ = '%';
    /* optional parameter: [0-9]+ '$' */

    /* '0' leading zero pad */
    if (info->pad != ' ')
	*be++ = info->pad;
    /* zero or more flag characters */
    if (info->left)
	*be++ = '-';
    if (info->showsign)
	*be++ = '+';
    if (info->space)
	*be++ = ' ';
    if (info->alt)
	*be++ = '#';
    if (info->group)
	*be++ = '\'';

    if (info->user == 0) {
 	/* optional minimum field width */
	if (info->width) {
	    (void) itoa(info->width, be, 10);
	    be += strlen(be);
	}
    } else {
	*be++ = '*';
    }

    /* optional precision */
    if (info->prec >= 0) {
	*be++ = '.';
	(void) itoa(info->prec, be, 10);
	be += strlen(be);
    }

    /* optional type modifier */
    if (info->user) {
#if	USE_EXAMPLE_XMM
	if (info->user & mod_v4f)
	    be = stpcpy(be, "v4f");
	if (info->user & mod_v2d)
	    be = stpcpy(be, "v2d");
	if (info->user & mod_v16i)
	    be = stpcpy(be, "v16i");
	if (info->user & mod_v8i)
	    be = stpcpy(be, "v8i");
	if (info->user & mod_v4i)
	    be = stpcpy(be, "v4i");
	if (info->user & mod_v2i)
	    be = stpcpy(be, "v2i");
#endif
#if	USE_EXAMPLE_HEX
	if (info->user & mod_phh)
	    be = stpcpy(be, "phh");
#ifdef	NOTYET
	if (info->user & mod_hC)
	    be = stpcpy(be, "hC");
	if (info->user & mod_hD)
	    be = stpcpy(be, "hD");
	if (info->user & mod_hN)
	    be = stpcpy(be, "hN");
#endif
#endif
    } else {
	if (info->is_long_double)
	    *be++ = 'L';	/* XXX 'll' instead? */
	else if (info->is_long)
	    *be++ = 'l';
	else if (info->is_short)
	    *be++ = 'h';
	else if (info->is_char) {
	    *be++ = 'h';
	    *be++ = 'h';
	}

	/* conversion character */
	switch (info->spec) {
	case 'c':
	case 's':
	    if (info->wide)
		*be++ = 'l';
	    break;
	}
    }
    *be++ = info->spec;
    *be = '\0';

    return b;
}

static void dumpPrintfInfo(const char *msg, const struct printf_info * info)
{
    char b[64];

    fprintf(stderr, "    %s \"%s\"\n", msg, info2fmt(info, b));

    if (info->extra || info->i18n)
	fprintf(stderr, "flags: %s%s\n",
		(info->extra		? " EXTRA " : ""),
		(info->i18n		? " I18N " : ""));
#ifdef	DYING
    if (info->user) {
	fprintf(stderr, " user: %d", (unsigned)info->user);
#if	USE_EXAMPLE_XMM
	if (info->user & mod_v4f)
	    fprintf(stderr, " %s", "v4f");
	if (info->user & mod_v2d)
	    fprintf(stderr, " %s", "v2d");
	if (info->user & mod_v16i)
	    fprintf(stderr, " %s", "v16i");
	if (info->user & mod_v8i)
	    fprintf(stderr, " %s", "v8i");
	if (info->user & mod_v4i)
	    fprintf(stderr, " %s", "v4i");
	if (info->user & mod_v2i)
	    fprintf(stderr, " %s", "v2i");
#endif
#if	USE_EXAMPLE_HEX
	if (info->user & mod_phh)
	    fprintf(stderr, " %s", "phh");
#ifdef	NOTYET
	if (info->user & mod_hC)
	    fprintf(stderr, " %s", "hC");
	if (info->user & mod_hD)
	    fprintf(stderr, " %s", "hD");
	if (info->user & mod_hN)
	    fprintf(stderr, " %s", "hN");
#endif
#endif
	fprintf(stderr, "\n");
    }
#endif	/* DYING */
}

#ifdef	REFERENCE
/*
--- http://en.cppreference.com/w/cpp/io/c/fprintf

    Formats:
	c	character
	s	string
	d,i	integer
	o	octal
	x,X	hex
	u	unsigned
	f,F	float decimal
	e,E	float decimal exponent
	a,A	float hex
	g,G	float decimal or decimal exponent
	n	no.chars written
	p	pointer

    Lengths:
	hh
	h
	l
	ll
	L	long double
	z	size_t
	j	intmax_t
	t	ptrdiff_t
	I	ptrdiff_t or size_t
	I32	int32_t
	I64	int64_t
	q	int64_t

--- https://www.kernel.org/doc/Documentation/printk-formats.txt
--- http://lxr.free-electrons.com/source/lib/test_printf.c
    Linux kernel:
	%pF	versatile_init + 0x0 / 0x110
	%pf	versatile_init
	%pS	versatile_init + 0x0 / 0x110
	%pSR	versatile_init + 0x9 / 0x110
			(with __builtin_extract_return_addr()translation)
	%ps	versatile_init
	%pB	prev_fn_of_versatile_init + 0x88 / 0x88
	%pK	0x01234567 or 0x0123456789abcdef
	%pr	[mem 0x60000000 - 0x6fffffff flags 0x2200] or
		[mem 0x0000000060000000 - 0x000000006fffffff flags 0x2200]
	%pR	[mem 0x60000000 - 0x6fffffff pref] or
		[mem 0x0000000060000000 - 0x000000006fffffff pref]

	%pa[p]	0x01234567 or 0x0123456789abcdef
	%pad	0x01234567 or 0x0123456789abcdef

	%*pE[achnops]
		%*pE	"\eb \C\a" \ 220 \ r] "
		%*pEhp	" \ x1bb \ C \ x07 "\x90\x0d]"
		%*pEa	"\e\142\040\\\103\a\042\220\r\135"
		a - ESCAPE_ANY
		c - ESCAPE_SPECIAL
		h - ESCAPE_HEX
		n - ESCAPE_NULL
		o - ESCAPE_OCTAL
		p - ESCAPE_NP
		s - ESCAPE_SPACE
	%*ph	00 01 02 ... 3f
	%*phC	00:01:02:...:3f
	%*phD	00 - 01 - 02 - ... -3f
	%*phN	000102 ... 3f
	%pM	00:01:02:03:04:05
	%pMR	05:04:03:02:01:00
	%pMF	00 - 01 - 02 - 03 - 04 - 05
	%pm	000102030405
	%pmR	050403020100

	%p[Ii] 4[hnbl]
		%pI4	1.2.3.4
		%pi4	001.002.003.004

		%pI6	0001:0002:0003:0004:0005:0006:0007:0008
		%pi6	00010002000300040005000600070008
		%pI6c	1:2:3:4:5:6:7:8

	%p[Ii]	S[pfschnbl]
		%pIS	1.2.3.4 or 0001:0002:0003:0004:0005:0006:0007:0008
		%piS	001.002.003.004 or 00010002000300040005000600070008
		%pISc	1.2.3.4 or 1:2:3:4:5:6:7:8
		%pISpc	1.2.3.4:12345 or [1:2:3:4:5:6:7:8]:12345

		%pISfc	1.2.3.4 or[1:2:3:4:5:6:7:8] / 123456789
		%pISsc	1.2.3.4 or[1:2:3:4:5:6:7:8] % 1234567890
		%pISpfc	1.2.3.4:12345 or[1:2:3:4:5:6:7:8]:12345 / 123456789
	%pUb	00010203-0405-0607-0809-0a0b0c0d0e0f
	%pUB	00010203-0405-0607-0809-0A0B0C0D0E0F
	%pUl	03020100-0504-0706-0809-0a0b0c0e0e0f
	%pUL	03020100-0504-0706-0809-0A0B0C0E0E0F
	%pd	{ , 2, 3, 4}
	%pD	{ , 2, 3, 4}

	%pg	sda, sda1 or loop0p1
	%pV
	%pC	pll1
	%pCn	pll1
	%pCr	1560000000
	%*pb	0779
	%*pbl	0, 3 - 6, 8 - 10
	%pGp	referenced | uptodate | lru | active | private
	%pGg	GFP_USER | GFP_DMA32 | GFP_NOWARN
	%pGv	read | exec | mayread | maywrite | mayexec | denywrite
	%pNF	0x000000000000c000
 */
#endif				/* REFERENCE */

static const char *atype[] = {
  "INT",	/* PA_INT	int */
  "CHAR",	/* PA_CHAR	int, cast to char */
  "WCHAR",	/* PA_WCHAR	wide char */
  "STRING",	/* PA_STRING	const char *, a '\0'-terminated string */
  "WSTRING",	/* PA_WSTRING	const wchar_t *, wide character string */
  "POINTER",	/* PA_POINTER	void * */
  "FLOAT",	/* PA_FLOAT	float */
  "DOUBLE",	/* PA_DOUBLE	double */
  "TYPE(8)",	/* PA_LAST */
  "TYPE(9)",
  "TYPE(10)",
  "TYPE(11)",
  "TYPE(12)",
  "TYPE(13)",
  "TYPE(14)",
  "TYPE(15)",
  "TYPE(16)",
  "TYPE(17)",
  "TYPE(18)",
  "TYPE(19)",
  "TYPE(20)",
  "TYPE(21)",
  "TYPE(22)",
  "TYPE(23)",
  "TYPE(24)",
  "TYPE(25)",
  "TYPE(26)",
  "TYPE(27)",
  "TYPE(28)",
  "TYPE(29)",
  "TYPE(30)",
  "TYPE(31)",
};

/*==============================================================*/
/* https://unixdev.ru/?p=47 */
#if	USE_EXAMPLE_IPADDR
static int print_ipv4_address(FILE * fp,
			      const struct printf_info *info,
			      const void *const *args)
{
    char s[INET_ADDRSTRLEN];
    const struct in_addr *addr;
    int ret = 1;
int w = info->width;
if (info->left) w = -w;

if (_tpf_debug) {
fprintf(stderr, "--> %s(%p, %p, %p)\n", __FUNCTION__, fp, info, args);
dumpPrintfInfo(__FUNCTION__, info);
}
    addr = *((const struct in_addr **) (args[0]));

    fprintf(fp, "%*s", w, inet_ntop(AF_INET, addr, s, INET_ADDRSTRLEN));
    fflush(fp);

if (_tpf_debug)
fprintf(stderr, "<-- %s(%p, %p, %p) ret %d\n", __FUNCTION__, fp, info, args, ret);
    return ret;
}

static int print_ipv6_address(FILE * fp, const struct printf_info *info,
			      const void *const *args)
{
    char s[INET6_ADDRSTRLEN];
    const struct in6_addr *addr;
    int ret = 1;
int w = info->width;
if (info->left) w = -w;

if (_tpf_debug) {
fprintf(stderr, "--> %s(%p, %p, %p)\n", __FUNCTION__, fp, info, args);
dumpPrintfInfo(__FUNCTION__, info);
}
    addr = *((const struct in6_addr **) (args[0]));

    fprintf(fp, "%*s", w, inet_ntop(AF_INET6, addr, s, INET6_ADDRSTRLEN));
    fflush(fp);

if (_tpf_debug)
fprintf(stderr, "<-- %s(%p, %p, %p) ret %d\n", __FUNCTION__, fp, info, args, ret);
    return ret;
}

static int print_address_size_arginfo(const struct printf_info *info, size_t n,
			      int *argtypes, int *sizes)
{
    int ret = 1;
if (_tpf_debug) {
fprintf(stderr, "--> %s(%p, %u)\n", __FUNCTION__, info, (unsigned)n);
dumpPrintfInfo(__FUNCTION__, info);
}
    if (n > 0) {
	argtypes[0] = PA_POINTER;
    }
if (_tpf_debug)
fprintf(stderr, "<-- %s(%p, %u) ret %d %s *sizes %d\n", __FUNCTION__, info, (unsigned)n, ret, atype[*argtypes&0x1f], *sizes);
    return ret;
}
#endif				/* USE_EXAMPLE_IPADDR */

/*==============================================================*/
/* http://plindenbaum.blogspot.com/2011/09/customizing-c-printf-to-print-fasta.html */

#if	USE_EXAMPLE_FASTA
typedef struct fasta_t {
    char *name;
    char *sequence;
} Fasta;

#define FPUTC(C) ret += (fputc(C,fp) == EOF ? 0 : 1)

/** handler for printf called by register_printf_function */
static int printf_fasta_handler(FILE * fp,
				const struct printf_info *info,
				const void *const *args)
{
    int n = 0;
    int ret = 0;
    char *p;
    /* return the number of character writtern */
    int lline = (info->prec < 1 ? 60 : info->prec);
    const Fasta *fasta = *((const Fasta **) (args[0]));

if (_tpf_debug) {
fprintf(stderr, "--> %s(%p, %p, %p)\n", __FUNCTION__, fp, info, args);
dumpPrintfInfo(__FUNCTION__, info);
}
    if (fasta->name != NULL) {
	FPUTC('>');
	ret += fputs(fasta->name, fp);
    } else {
	FPUTC('>');
    }

    if (fasta->sequence != NULL) {
	p = (char *) fasta->sequence;
	while (p[n] != 0 && (info->width < 1 || n < info->width)) {
	    if (n % lline == 0) {
		FPUTC('\n');
	    }
	    FPUTC(p[n]);
	    ++n;
	}
    }
    FPUTC('\n');
if (_tpf_debug)
fprintf(stderr, "<-- %s(%p, %p, %p) ret %d\n", __FUNCTION__, fp, info, args, ret);
    return ret;
}

static int printf_fasta_arginfo(const struct printf_info *info,
				size_t n, int *argtypes, int *sizes)
{
    int ret = 1;
if (_tpf_debug) {
fprintf(stderr, "--> %s(%p, %u)\n", __FUNCTION__, info, (unsigned)n);
dumpPrintfInfo(__FUNCTION__, info);
}
    /*
     * We always take exactly one argumentl this is a pointer to the structure.
     */
    if (n > 0)
	argtypes[0] = PA_POINTER;
if (_tpf_debug)
fprintf(stderr, "<-- %s(%p, %u) ret %d %s *sizes %d\n", __FUNCTION__, info, (unsigned)n, ret, atype[*argtypes&0x1f], *sizes);
    return ret;
}
#endif				/* USE_EXAMPLE_FASTA */

/*==============================================================*/
/* https://github.com/libdfp/libdfp/blob/master/old_printf_dfp.c */

/*==============================================================*/
/* https://lists.freedesktop.org/archives/gstreamer-commits/2009-May/034572.html */

/*==============================================================*/
/* https://oss.oracle.com/pipermail/ocfs2-tools-devel/2012-February/004067.html */

/*==============================================================*/

#if	USE_EXAMPLE_WIDGET
typedef struct {
    char *name;
} Widget;

static int
print_widget(FILE * fp,
	     const struct printf_info *info, const void *const *args)
{
    const Widget *w;
    char *buffer = NULL;
    int len;
    int ret = 0;

if (_tpf_debug) {
fprintf(stderr, "--> %s(%p, %p, %p)\n", __FUNCTION__, fp, info, args);
dumpPrintfInfo(__FUNCTION__, info);
}
    /* Format the output into a string. */
    w = *((const Widget **) (args[0]));
    len = asprintf(&buffer, "<Widget %p: %s>", w, w->name);
    if (len == -1)
	ret = -1;
    else {
	/* Pad to the minimum field width and print to the stream. */
	len = fprintf(fp, "%*s",
		  (info->left ? -info->width : info->width), buffer);
	fflush(fp);
    }

    /* Clean up and return. */
    if (buffer)
	free(buffer);
if (_tpf_debug)
fprintf(stderr, "<-- %s(%p, %p, %p) ret %d len %d\n", __FUNCTION__, fp, info, args, ret, len);
    return ret;
}

/**
 * @return		1 on success, -1 on cannot handle
 **/
static int
print_widget_size_arginfo(const struct printf_info *info, size_t n,
			  int *argtypes, int *sizes)
{
    int ret = 1;
if (_tpf_debug) {
fprintf(stderr, "--> %s(%p, %u)\n", __FUNCTION__, info, (unsigned)n);
dumpPrintfInfo(__FUNCTION__, info);
}
    /* We always take exactly one argument and this is a pointer to the
       structure.. */
    if (n > 0)
	argtypes[0] = PA_POINTER;
if (_tpf_debug)
fprintf(stderr, "<-- %s(%p, %u) ret %d %s *sizes %d\n", __FUNCTION__, info, (unsigned)n, ret, atype[*argtypes&0x1f], *sizes);
    return ret;
}
#endif				/* USE_EXAMPLE_WIDGET */

/*==============================================================*/
/* https://sourceware.org/bugzilla/attachment.cgi?id=3874 */
/* https://sourceware.org/ml/libc-help/2015-01/msg00005.html */
/* http://udrepper.livejournal.com */
/* https://www.redhat.com/archives/libguestfs/2009-November/msg00255.html */

#if	USE_EXAMPLE_XMM

static void xmm_va(void *mem, va_list * ap)
{
if (_tpf_debug)
fprintf(stderr, "--> %s(%p, %p)\n", __FUNCTION__, mem, ap);
    __m128 v = va_arg(*ap, __m128);
    memcpy(mem, &v, sizeof(v));
}

static int
xmm_ais(const struct printf_info *info, size_t n, int *argtypes, int *sizes)
{
    int ret = 1;
if (_tpf_debug) {
fprintf(stderr, "--> %s(%p, %u)\n", __FUNCTION__, info, (unsigned)n);
dumpPrintfInfo(__FUNCTION__, info);
}
    argtypes[0] = pa_xmm;
    sizes[0] = sizeof(__m128);
if (_tpf_debug)
fprintf(stderr, "<-- %s(%p, %u) ret %d %s *sizes %d\n", __FUNCTION__, info, (unsigned)n, ret, atype[*argtypes&0x1f], *sizes);
    return ret;
}

static int
xmm_printf_f(FILE * fp, const struct printf_info *info,
	     const void *const *args)
{
    int len = 0;
    int ret = 0;

if (_tpf_debug) {
fprintf(stderr, "--> %s(%p, %p, %p)\n", __FUNCTION__, fp, info, args);
dumpPrintfInfo(__FUNCTION__, info);
}
    if (info->user & mod_v4f) {
	union {
	    __m128 v;
	    float f[4];
	} u;
	memcpy(&u, *((void ***) args)[0], sizeof(__m128));
	char *buf;
	len = asprintf(&buf, "{%f, %f, %f, %f}", u.f[0], u.f[1], u.f[2],
		     u.f[3]);
	fputs(buf, fp);
	free(buf);
    } else if (info->user & mod_v2d) {
	union {
	    __m128d v;
	    double d[2];
	} u;
	memcpy(&u, *((void ***) args)[0], sizeof(__m128d));
	char *buf;
	len = asprintf(&buf, "{%f, %f}", u.d[0], u.d[1]);
	fputs(buf, fp);
	free(buf);
    } else
	ret = -2;

if (_tpf_debug)
fprintf(stderr, "<-- %s(%p, %p, %p) ret %d len %d\n", __FUNCTION__, fp, info, args, ret, len);
    return ret;
}

static int
xmm_printf_x(FILE * fp, const struct printf_info *info,
	     const void *const *args)
{
    int len = 0;
    int ret = 0;

if (_tpf_debug) {
fprintf(stderr, "--> %s(%p, %p, %p)\n", __FUNCTION__, fp, info, args);
dumpPrintfInfo(__FUNCTION__, info);
}
    if (info->user & mod_v16i) {
	union {
	    __m128i v;
	    uint8_t i[16];
	} u;
	memcpy(&u, *((void ***) args)[0], sizeof(__m128i));
	char *buf;
	len = asprintf(&buf,
		     "{%" PRIx8 ", %" PRIx8 ", %" PRIx8 ", %" PRIx8 ", %"
		     PRIx8 ", %" PRIx8 ", %" PRIx8 ", %" PRIx8 ", " "%"
		     PRIx8 ", %" PRIx8 ", %" PRIx8 ", %" PRIx8 ", %" PRIx8
		     ", %" PRIx8 ", %" PRIx8 ", %" PRIx8 "}", u.i[0],
		     u.i[1], u.i[2], u.i[3], u.i[4], u.i[5], u.i[6],
		     u.i[7], u.i[8], u.i[9], u.i[10], u.i[11], u.i[12],
		     u.i[13], u.i[14], u.i[15]);
	fputs(buf, fp);
	free(buf);
    } else if (info->user & mod_v8i) {
	union {
	    __m128i v;
	    uint16_t i[8];
	} u;
	memcpy(&u, *((void ***) args)[0], sizeof(__m128i));
	char *buf;
	len = asprintf(&buf,
		     "{%" PRIx16 ", %" PRIx16 ", %" PRIx16 ", %" PRIx16
		     ", %" PRIx16 ", %" PRIx16 ", %" PRIx16 ", %" PRIx16
		     "}", u.i[0], u.i[1], u.i[2], u.i[3], u.i[4], u.i[5],
		     u.i[6], u.i[7]);
	fputs(buf, fp);
	free(buf);
    } else if (info->user & mod_v4i) {
	union {
	    __m128i v;
	    uint32_t i[4];
	} u;
	memcpy(&u, *((void ***) args)[0], sizeof(__m128i));
	char *buf;
	len = asprintf(&buf,
		     "{%" PRIx32 ", %" PRIx32 ", %" PRIx32 ", %" PRIx32
		     "}", u.i[0], u.i[1], u.i[2], u.i[3]);
	fputs(buf, fp);
	free(buf);
    } else if (info->user & mod_v2i) {
	union {
	    __m128i v;
	    uint64_t i[2];
	} u;
	memcpy(&u, *((void ***) args)[0], sizeof(__m128i));
	char *buf;
	len = asprintf(&buf, "{%" PRIx64 ", %" PRIx64 "}", u.i[0], u.i[1]);
	fputs(buf, fp);
	free(buf);
    } else
	ret = -2;

if (_tpf_debug)
fprintf(stderr, "<-- %s(%p, %p, %p) ret %d len %d\n", __FUNCTION__, fp, info, args, ret, len);

    return ret;
}
#endif				/* USE_EXAMPLE_XMM */

/*==============================================================*/

#if	USE_EXAMPLE_HEX

static const char hexdigs[] = "0123456789ABCDEF";

typedef	union dataPtrs_u DataPtrs; 
union dataPtrs_u {
    void * ptr;
    uint8_t * ui8p;
    uint16_t * ui16p;
    uint32_t * ui32p;
    uint64_t * ui64p;
    const char * str;
    unsigned char * blob;
    const char ** argv;
};

static void hex_va(void *mem, va_list * ap)
{
    DataPtrs p;
    p.ptr = va_arg(*ap, void *);
if (_tpf_debug)
fprintf(stderr, "\tmemcpy(%p, %p, %u) p.ptr %p\n", mem, &p, (unsigned)(sizeof(p)), p.ptr);
    memcpy(mem, &p, sizeof(p));
}

static int
hex_ais(const struct printf_info *info, size_t n, int *argtypes, int *sizes)
{
    char b[64];
    int ret = 1;

    argtypes[0] = pa_hex;
    sizes[0] = sizeof(uint8_t *);	/* XXX HACK */
if (_tpf_debug)
fprintf(stderr, "<-- %s(\"%s\", %u) ret %d %s *sizes %d pad '%c' field %u/%u\n", __FUNCTION__, info2fmt(info, b), (unsigned)n, ret, atype[*argtypes&0x1f], *sizes, info->pad, info->width, info->prec);
    return ret;
}

static int
hex_printf_p(FILE * fp, const struct printf_info *info,
	     const void *const *args)
{
    char b[64];
    int len = 0;
    int ret = 0;

#ifdef	NOTYET
    if (info->user & mod_v4f) {
	DataPtrs u;
	memcpy(&u, *((void ***) args)[0], sizeof(__m128));
	char *buf;
	len = asprintf(&buf, "{%f, %f, %f, %f}", u.f[0], u.f[1], u.f[2],
		     u.f[3]);
	fputs(buf, fp);
	free(buf);
    } else if (info->user & mod_v2d) {
	DataPtrs u;
	memcpy(&u, *((void ***) args)[0], sizeof(__m128d));
	char *buf;
	len = asprintf(&buf, "{%f, %f}", u.d[0], u.d[1]);
	fputs(buf, fp);
	free(buf);
    } else
#endif
	ret = -2;

if (_tpf_debug)
fprintf(stderr, "<-- %s(\"%s\") ret %d len %d\n", __FUNCTION__, info2fmt(info, b), ret, len);
    return ret;
}

static int
hex_printf_dox(FILE * fp, const struct printf_info *info,
	     const void *const *args)
{
    char b[64];
    int len = 0;
    int ret = 0;

    if (info->user & mod_phh) {
	DataPtrs p;
	const char * pfx = NULL;
	int base;
	int ndigs;
	char pad;
	int ilen;

	memcpy(&p, *((void ***) args)[0], sizeof(p.ptr));

	pad = (info->pad == '0' ? '\0' : info->pad);
	switch (info->spec) {
	case 'd':
	    pfx = NULL;
	    base = 10;
	    ndigs = 3;
	    break;
	case 'o':
	    pfx = (info->alt ? "0" : NULL);
	    base = 8;
	    ndigs = 3;
	    break;
	case 'x':
	default:
	    pfx = (info->alt ? "0x" : NULL);
	    base = 16;
	    ndigs = 2;
	    break;
	}
	ilen = ndigs;
	if (info->alt) {
	    pad = ',';
	    if (pfx)
		ilen += strlen(pfx);
	}
	
	len = (pad
	    ? (ilen+1) * info->width - 1
	    : ilen * info->width );
	char *buf = malloc(len + 1);
	char *be = buf;
	for (int j = 0; j < info->width; j++) {
	    unsigned v = p.ui8p[j];
	    if (j > 0 && pad)
		*be++ = pad;
	    if (pfx)
		be = stpcpy(be, pfx);
	    switch (info->spec) {
	    case 'd':
		itoa(v, be, base);
		be += strlen(be);
		break;
	    case 'o':
		*be++ = hexdigs[(v >> 6) & 007];
		*be++ = hexdigs[(v >> 3) & 007];
		*be++ = hexdigs[(v     ) & 007];
		break;
	    case 'x':
		*be++ = hexdigs[(v >> 4) & 0x0f];
		*be++ = hexdigs[(v     ) & 0x0f];
		break;
	    }
	    *be = '\0';
	}
	fputs(buf, fp);
	free(buf);
    } else if (info->user & mod_hC) {
#ifdef	NOTYET
fprintf(stderr, "\t%s %p\n", "hC", *((void ***)args)[0]);
	union {
	    __m128i v;
	    uint16_t i[8];
	} u;
	memcpy(&u, *((void ***) args)[0], sizeof(__m128i));
	char *buf;
	len = asprintf(&buf,
		     "{%" PRIx16 ", %" PRIx16 ", %" PRIx16 ", %" PRIx16
		     ", %" PRIx16 ", %" PRIx16 ", %" PRIx16 ", %" PRIx16
		     "}", u.i[0], u.i[1], u.i[2], u.i[3], u.i[4], u.i[5],
		     u.i[6], u.i[7]);
	fputs(buf, fp);
	free(buf);
#endif
    } else if (info->user & mod_hD) {
#ifdef	NOTYET
fprintf(stderr, "\t%s %p\n", "hD", *((void ***)args)[0]);
	union {
	    __m128i v;
	    uint32_t i[4];
	} u;
	memcpy(&u, *((void ***) args)[0], sizeof(__m128i));
	char *buf;
	len = asprintf(&buf,
		     "{%" PRIx32 ", %" PRIx32 ", %" PRIx32 ", %" PRIx32
		     "}", u.i[0], u.i[1], u.i[2], u.i[3]);
	fputs(buf, fp);
	free(buf);
#endif
    } else if (info->user & mod_hN) {
#ifdef	NOTYET
fprintf(stderr, "\t%s %p\n", "hN", *((void ***)args)[0]);
	union {
	    __m128i v;
	    uint64_t i[2];
	} u;
	memcpy(&u, *((void ***) args)[0], sizeof(__m128i));
	char *buf;
	len = asprintf(&buf, "{%" PRIx64 ", %" PRIx64 "}", u.i[0], u.i[1]);
	fputs(buf, fp);
	free(buf);
#endif
    } else
	ret = -2;

if (_tpf_debug)
fprintf(stderr, "<-- %s(\"%s\") ret %d len %d\n", __FUNCTION__, info2fmt(info, b), ret, len);

    return ret;
}
#endif				/* USE_EXAMPLE_HEX */

/*==============================================================*/

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wformat"
#pragma GCC diagnostic ignored "-Wformat-extra-args"

int main(void)
{
    FILE *fp = stderr;

/*==============================================================*/

#if	USE_EXAMPLE_IPADDR
fprintf(stderr, "==> IPADDR\n");
    struct in6_addr ip6_addr;
    struct in_addr ip4_addr;

    register_printf_specifier('P', print_ipv4_address, print_address_size_arginfo);
    register_printf_specifier('W', print_ipv6_address, print_address_size_arginfo);

    inet_pton(AF_INET6, "2001::1", &ip6_addr);
    inet_pton(AF_INET, "10.10.0.1", &ip4_addr);

    fprintf(fp, "IPv4 addr: %040.5P\nIPv6 addr: %0-40W\n", &ip4_addr, &ip6_addr);
    fflush(fp);
fprintf(stderr, "<== IPADDR\n");
#endif				/* USE_EXAMPLE_IPADDR */

/*==============================================================*/

#if	USE_EXAMPLE_FASTA

#define APPEND(ptr,len,c) \
    {	ptr = realloc(ptr, sizeof(char)*((len)+2)); \
	if (ptr == NULL) {\
	    fputs("Out of memory\n", stderr); \
	    exit(EXIT_FAILURE); \
	} \
	ptr[(len)] = c; \
	ptr[(len)+1] = 0; \
	len++; \
    }

fprintf(stderr, "==> FASTA\n");
    size_t len_name = 0;
    size_t len_seq = 0;
    Fasta record;
    int c;

    /* register a new output conversion */
    if (register_printf_specifier('W',
				  printf_fasta_handler,
				  printf_fasta_arginfo) != 0)
    {
	fputs("Cannot register print function", stderr);
	return EXIT_FAILURE;
    }
    record.name = NULL;
    record.sequence = NULL;
    for (;;) {
	c = fgetc(stdin);
	if (c == EOF || c == '>') {
	    if (len_seq > 0) {
		fprintf(fp, "%80W\n", &record);		fflush(fp);
		fprintf(fp, "%200.90W\n", &record);	fflush(fp);
		fprintf(fp, "%200W\n", &record);	fflush(fp);
		fprintf(fp, "%.80W\n", &record);	fflush(fp);
	    }
	    if (c == EOF)
		break;
	    len_seq = 0;
	    len_name = 0;
	    while ((c = fgetc(stdin)) != EOF && c != '\n') {
		APPEND(record.name, len_name, c);
	    }
	} else if (!isspace(c)) {
	    APPEND(record.sequence, len_seq, c);
	}
    }
    free(record.name);
    free(record.sequence);
    fflush(stdout);
fprintf(stderr, "<== FASTA\n");
#endif				/* USE_EXAMPLE_FASTA */

/*==============================================================*/

#if	USE_EXAMPLE_WIDGET
fprintf(stderr, "==> WIDGET\n");
    /* Make a widget to print. */
    Widget mywidget;
    mywidget.name = "mywidget";

    /* Register the print function for widgets. */
    register_printf_specifier('W', print_widget,
			      print_widget_size_arginfo);

    /* Now print the widget. */
    fprintf(fp, "|%W|\n", &mywidget);	fflush(fp);
    fprintf(fp, "|%35W|\n", &mywidget);	fflush(fp);
    fprintf(fp, "|%-35W|\n", &mywidget);fflush(fp);
fprintf(stderr, "<== WIDGET\n");
#endif				/* USE_EXAMPLE_WIDGET */

/*==============================================================*/

#if	USE_EXAMPLE_XMM

fprintf(stderr, "==> XMM\n");

    pa_xmm = register_printf_type(xmm_va);

    mod_v4f = register_printf_modifier(L"v4f");
    mod_v2d = register_printf_modifier(L"v2d");
    mod_v16i = register_printf_modifier(L"v16i");
    mod_v8i = register_printf_modifier(L"v8i");
    mod_v4i = register_printf_modifier(L"v4i");
    mod_v2i = register_printf_modifier(L"v2i");

    register_printf_specifier('f', xmm_printf_f, xmm_ais);
    register_printf_specifier('x', xmm_printf_x, xmm_ais);

    __m128 f = _mm_set_ps(1.0, 2.0, 3.0, 4.0);
    __m128d d = _mm_set_pd(1.0, 2.0);
    __m128i i =
	_mm_set_epi8(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
		     16);

    fprintf(fp, "%f\n", 1.0f);	fflush(fp);
    fprintf(fp, "%f\n", 2.0);	fflush(fp);
    fprintf(fp, "%v4ff\n", f);	fflush(fp);
    fprintf(fp, "%v2df\n", d);	fflush(fp);
    fprintf(fp, "%x\n", 1);	fflush(fp);
    fprintf(fp, "%v16ix\n", i);	fflush(fp);
    fprintf(fp, "%v8ix\n", i);	fflush(fp);
    fprintf(fp, "%v4ix\n", i);	fflush(fp);
    fprintf(fp, "%v2ix\n", i);	fflush(fp);
fprintf(stderr, "<== XMM\n");
#endif				/* USE_EXAMPLE_XMM */

/*==============================================================*/

#if	USE_EXAMPLE_HEX

fprintf(stderr, "==> HEX\n");

    pa_hex = register_printf_type(hex_va);

    mod_phh = register_printf_modifier(L"phh");
#ifdef	NOTYET
    mod_hC = register_printf_modifier(L"hC");
    mod_hD = register_printf_modifier(L"hD");
    mod_hN = register_printf_modifier(L"hN");
#endif

    if (!_tpf_debug) {
	register_printf_specifier('p', hex_printf_p, hex_ais);
	register_printf_specifier('d', hex_printf_dox, hex_ais);
    }
    register_printf_specifier('o', hex_printf_dox, hex_ais);
    register_printf_specifier('x', hex_printf_dox, hex_ais);

    uint8_t b[] = {
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
	0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
	0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
    };

fprintf(stderr, "========== %p[%u]\n", b, sizeof(b));

    if (!_tpf_debug) {
#ifdef	BUGGY
	fprintf(fp, "%%*phhp[%u]: %*phhp[%u]\n", sizeof(b)/4, b, sizeof(b));	fflush(fp);
#endif
	fprintf(fp, "   %%*phhp: %*phhp\n", sizeof(b)/4, b);	fflush(fp);
#ifdef	NOTYET
	fprintf(fp, "%*hp\n", sizeof(b), b);	fflush(fp);
	fprintf(fp, "%*hCp\n", sizeof(b), b);	fflush(fp);
	fprintf(fp, "%*hDp\n", sizeof(b), b);	fflush(fp);
	fprintf(fp, "%*hNp\n", sizeof(b), b);	fflush(fp);

	fprintf(fp, "%#lx\n", (unsigned long)b);	fflush(fp);
#endif
	fprintf(fp, "   %%*phhd: %*phhd\n", sizeof(b)/4, b);	fflush(fp);
	fprintf(fp, "%%*.16phhd: %*.16phhd\n", sizeof(b)/4, b);	fflush(fp);
	fprintf(fp, "  %% *phhd: % *phhd\n", sizeof(b)/4, b);	fflush(fp);
#ifdef	BUGGY	/* XXX same as %u */
	fprintf(fp, "  %%* phhd: %* phhd\n", sizeof(b)/4, b);	fflush(fp);
#endif
	fprintf(fp, "  %%0*phhd: %0*phhd\n", sizeof(b)/4, b);	fflush(fp);
	fprintf(fp, "  %%#*phhd: %#*phhd\n", sizeof(b)/4, b);	fflush(fp);
	fprintf(fp, "  %%'*phhd: %'*phhd\n", sizeof(b)/4, b);	fflush(fp);
    }
    fprintf(fp, "   %%*phhx: %*phhx\n", sizeof(b)/4, b);	fflush(fp);
    fprintf(fp, "%%*.16phhx: %*.16phhx\n", sizeof(b)/4, b);	fflush(fp);
    fprintf(fp, "  %% *phhx: % *phhx\n", sizeof(b)/4, b);	fflush(fp);
#ifdef	BUGGY	/* XXX same as %u */
    fprintf(fp, "  %%* phhx: %* phhx\n", sizeof(b)/4, b);	fflush(fp);
#endif
    fprintf(fp, "  %%0*phhx: %0*phhx\n", sizeof(b)/4, b);	fflush(fp);
    fprintf(fp, "  %%#*phhx: %#*phhx\n", sizeof(b)/4, b);	fflush(fp);
    fprintf(fp, "  %%'*phhx: %'*phhx\n", sizeof(b)/4, b);	fflush(fp);

    fprintf(fp, "   %%*phho: %*phho\n", sizeof(b)/4, b);	fflush(fp);
    fprintf(fp, "  %%#*phho: %#*phho\n", sizeof(b)/4, b);	fflush(fp);
#ifdef	NOTYET
    fprintf(fp, "%*hCx\n", sizeof(b), b);	fflush(fp);
    fprintf(fp, "%*hDx\n", sizeof(b), b);	fflush(fp);
    fprintf(fp, "%*hNx\n", sizeof(b), b);	fflush(fp);
#endif

fprintf(stderr, "<== HEX\n");

#endif				/* USE_EXAMPLE_XMM */

/*==============================================================*/

    return 0;
}

#pragma GCC diagnostic pop
